import click

import ckan.plugins.toolkit as tk
from ckan import model
from ckan.lib.search import clear, query_for


@click.group(short_help="Extra tools for managing search-index")
def search_index():
    pass


@search_index.command()
@click.option("-y", "--no-confirm", is_flag=True)
@click.pass_context
def clear_missing(ctx, no_confirm: bool):
    """Drop packages that are only in search index but not in DB."""
    q = query_for("package")
    with ctx.meta["flask_app"].test_request_context():
        limit = tk.get_action("package_search")({"ignore_auth": True}, {"rows": 0})["count"]
    with click.progressbar(q.get_all_entity_ids(limit)) as bar:
        query = model.Session.query(model.Package.id)
        ids = {id for id in bar if not model.Session.query(query.filter_by(id=id).exists()).scalar()}

    if not ids:
        click.secho("No missing packages detected", fg="green")
        return

    click.echo("Following packages are missing from the database:")
    for id in ids:
        click.echo("\t" + id)

    if not no_confirm:
        abort = not click.confirm(
            "Do you want to remove these packages from the search index?",
        )
        if abort:
            raise click.Abort

    with click.progressbar(ids) as bar:
        for id_ in bar:
            clear(id_)



"""Tests for IPackageController implementation.

Here you should verify that any modifications from intterface implementations
work as expected.
"""

from __future__ import annotations

from typing import Any

import pytest


@pytest.mark.usefixtures("with_plugins", "clean_db")
def test_after_show(package: dict[str, Any]):
    assert package["fake"] == 42



from __future__ import annotations

from datetime import datetime, timedelta
from operator import itemgetter
from typing import TYPE_CHECKING

import pytest

from ckanext.toolbelt.utils.tracking import DateTracker, Tracker

if TYPE_CHECKING:
    from faker import Faker
    from freezegun.api import FrozenDateTimeFactory


@pytest.fixture
def tracker_factory():
    return Tracker


@pytest.fixture
def tracker(faker: Faker, tracker_factory: type[Tracker]):
    return tracker_factory(faker.word())


@pytest.mark.usefixtures("clean_redis")
class TestAllTrackers:
    @pytest.fixture(params=[Tracker, DateTracker])
    def tracker_factory(self, request: pytest.FixtureRequest):
        return request.param

    def test_hits(self, tracker: Tracker, faker: Faker):
        """Each event tracked separately."""
        event = faker.word()
        another_event = faker.word()
        n = 10

        for idx in range(n):
            tracker.hit(event)
            if idx % 2:
                tracker.hit(another_event)

        assert tracker.score(event) == n
        assert tracker.score(another_event) == n // 2

    def test_namespaces(self, tracker_factory: type[Tracker], faker: Faker):
        """Trackers created with the same name share the data."""
        event = faker.word()
        name = faker.word()

        global_count, named_count = faker.pyint(), faker.pyint()

        global_track = tracker_factory()
        named_track = tracker_factory(name)

        global_track.hit(event, global_count)
        named_track.hit(event, named_count)

        assert tracker_factory().score(event) == global_count
        assert tracker_factory(name).score(event) == named_count

    def test_throttling(self, tracker_factory: type[Tracker], faker: Faker):
        """Throttling ignores duplicated events for certain amount of time."""
        event = faker.word()
        track = tracker_factory(throttling_time=10)
        track.hit(event)
        track.hit(event)
        assert track.score(event) == 1

        track.unthrottle(track.hash(event))

        track.hit(event)
        track.hit(event)
        assert track.score(event) == 2

    def test_ignoring(self, tracker_factory: type[Tracker], faker: Faker):
        """Ignored events are not tracked at all."""
        event = faker.word()
        track = tracker_factory()
        hash = track.hash(event)
        track.ignore(hash)

        track.hit(event)
        track.hit(event)
        assert track.score(event) == 0

        track.unignore(hash)

        track.hit(event)
        track.hit(event)
        assert track.score(event) == 2

    def test_translations(self, tracker: Tracker, faker: Faker):
        """Tracker can restore event name from it's hash."""
        event = faker.word()
        tracker.hit(event)
        hash = tracker.hash(event)

        assert tracker.translate(hash) == event

    def test_drop(self, tracker: Tracker, faker: Faker):
        """Event scores can be removed."""
        event = faker.word()
        tracker.hit(event)
        tracker.drop(event)
        assert tracker.score(event) == 0

    def test_reset(self, tracker: Tracker, faker: Faker):
        """Tracker data can be removed."""
        for _ in range(10):
            tracker.hit(faker.word())
        tracker.reset()

        assert not tracker.snapshot()

    def test_most_common(self, tracker: Tracker, faker: Faker):
        """Most common items ordered by score."""
        first_event = faker.word()
        second_event = faker.word()
        third_event = faker.word()

        tracker.hit(first_event, 100)
        tracker.hit(second_event, 50)
        tracker.hit(third_event, 200)

        if isinstance(tracker, DateTracker):
            tracker.refresh()

        assert list(tracker.most_common(1)) == [{"event": third_event, "score": 200}]
        assert list(tracker.most_common(2)) == [
            {"event": third_event, "score": 200},
            {"event": first_event, "score": 100},
        ]
        assert list(tracker.most_common(3)) == [
            {"event": third_event, "score": 200},
            {"event": first_event, "score": 100},
            {"event": second_event, "score": 50},
        ]

    def test_obsoletion_period(
        self,
        tracker: Tracker,
        faker: Faker,
        freezer: FrozenDateTimeFactory,
    ):
        """Most common items ordered by score."""
        if not isinstance(tracker, DateTracker):
            return

        first_event = faker.word()
        second_event = faker.word()

        tracker.hit(first_event, 100)
        freezer.move_to(timedelta(seconds=tracker.max_age // 2))
        tracker.hit(second_event, 50)

        tracker.refresh()
        assert list(tracker.most_common(2)) == [
            {"event": first_event, "score": 100},
            {"event": second_event, "score": 50},
        ]

        freezer.move_to(timedelta(seconds=tracker.max_age // 2))
        tracker.refresh()
        assert list(tracker.most_common(2)) == [{"event": second_event, "score": 50}]

        freezer.move_to(timedelta(seconds=tracker.max_age // 2))
        tracker.refresh()
        assert list(tracker.most_common(2)) == []

    def test_score(
        self,
        tracker: Tracker,
        faker: Faker,
        freezer: FrozenDateTimeFactory,
    ):
        """Most common items ordered by score."""
        event = faker.word()

        tracker.hit(event, 100)
        if isinstance(tracker, DateTracker):
            freezer.move_to(timedelta(seconds=tracker.obsoletion_period * 3))
        tracker.hit(event, 100)

        if isinstance(tracker, DateTracker):
            assert tracker.score(event) == 125
        else:
            assert tracker.score(event) == 200


@pytest.mark.usefixtures("clean_redis")
class TestTracker:
    def test_snapshot(self, tracker: Tracker, faker: Faker):
        """Tracker data has expected snapshot format."""
        first_event = faker.word()
        second_event = faker.word()
        first_count = faker.pyint()
        second_count = faker.pyint()
        tracker.hit(first_event, first_count)
        tracker.hit(second_event, second_count)
        assert sorted(tracker.snapshot(), key=itemgetter("event")) == sorted(
            [
                {"event": first_event, "score": first_count},
                {"event": second_event, "score": second_count},
            ],
            key=itemgetter("event"),
        )

    def test_restore(self, tracker: Tracker, faker: Faker):
        """Tracker data can be restored from snapshot format."""
        first_event = faker.word()
        second_event = faker.word()
        first_count = faker.pyint()
        second_count = faker.pyint()
        tracker.restore(
            [
                {"event": first_event, "score": first_count},
                {"event": second_event, "score": second_count},
            ],
        )
        assert tracker.score(first_event) == first_count
        assert tracker.score(second_event) == second_count


@pytest.mark.usefixtures("clean_redis")
class TestDateTracker:
    @pytest.fixture
    def tracker_factory(self):
        return DateTracker

    def test_moment_interactions(self, tracker, faker):
        """Past scores can be updated and shown."""
        date = faker.date_time_between(end_date="-2d")
        event = faker.word()
        tracker.hit(event, moment=date)

        assert tracker.moment_score(event) == 0
        assert tracker.moment_score(event, moment=date) == 1

    def test_snapshot(self, tracker: DateTracker, faker: Faker):
        """Tracker data has expected snapshot format."""
        first_event = faker.word()
        second_event = faker.word()
        first_count = faker.pyint()
        second_count = faker.pyint()
        tracker.hit(first_event, first_count)
        tracker.hit(second_event, second_count)

        date = datetime.strptime(
            tracker.format_date_stem(tracker.now()),
            tracker.date_format,
        )

        assert sorted(tracker.snapshot(), key=itemgetter("event")) == sorted(
            [
                {
                    "event": first_event,
                    "records": [{"date": date.isoformat(), "score": first_count}],
                },
                {
                    "event": second_event,
                    "records": [{"date": date.isoformat(), "score": second_count}],
                },
            ],
            key=itemgetter("event"),
        )

    def test_restore(self, tracker: DateTracker, faker: Faker):
        """Tracker data can be restored from snapshot format."""
        first_event = faker.word()
        second_event = faker.word()
        first_count = faker.pyint()
        second_count = faker.pyint()
        date = datetime.strptime(
            tracker.format_date_stem(tracker.now()),
            tracker.date_format,
        )

        tracker.restore(
            [
                {
                    "event": first_event,
                    "records": [{"date": date.isoformat(), "score": first_count}],
                },
                {
                    "event": second_event,
                    "records": [{"date": date.isoformat(), "score": second_count}],
                },
            ],
        )
        assert tracker.score(first_event) == first_count
        assert tracker.score(second_event) == second_count



from __future__ import annotations

from typing import Any

import pytest
from faker import Faker
from playwright.sync_api import Page, expect

from ckan import types
from ckan.tests.helpers import call_action  # pyright: ignore[reportUnknownVariableType]

from ckanext.toolbelt.tests.e2e_base._locators import ElementLocator

__all__ = [
    "Login",
    "Logout",
    "Profile",
    "StandardUser",
]


class Login(ElementLocator):
    @pytest.mark.ckan_standard("login")
    def test_login_normal(self, page: Page, user_factory: types.TestFactory, faker: Faker, milestone_screenshot: Any):
        """Test normal login flow.

        Given an existing user
        When they go to the login page
        And fill in their username and password
        And submit the form
        Then they are redirected to the dashboard
        And their display name is shown in the header
        """
        password = faker.password()
        user = user_factory(password=password)

        page.goto("/")

        self.locate_login_link(page).click()
        expect(page).to_have_url("/user/login")

        page.get_by_label("username").fill(user["name"])
        page.get_by_label("password").fill(password)

        milestone_screenshot("form_with_credentials")

        self.locate_login_button(page).click()
        expect(page).to_have_url("/dashboard/datasets")

        profile_link = self.locate_profile_link(page, user)
        expect(profile_link).to_be_visible()

        milestone_screenshot("after_login_redirect")

    @pytest.mark.ckan_standard("login")
    def test_login_authenticated(self, milestone_screenshot: Any, login: Any, user: dict[str, Any], page: Page):
        """Authenticated user cannot access login page again.

        Given an authenticated user
        When they go to the login page
        Then they see a message that they are already logged in
        And the username and password fields are disabled
        """
        login(user["name"])

        page.goto("/user/login")
        expect(self.locate_login_link(page)).not_to_be_attached()

        expect(self.locate_existing_login_session_alert(page)).to_be_visible()
        milestone_screenshot("existing_session_alert")


class Logout(ElementLocator):
    @pytest.mark.ckan_standard("logout")
    def test_logout_normal(self, milestone_screenshot: Any, user: dict[str, Any], page: Page, login: Any):
        """Test normal logout flow.

        Given an authenticated user
        When they click the logout link
        Then they are redirected to the post-logout page
        And they see a message that they have been logged out
        And the login link is visible in the header
        """
        login(user)
        page.goto("/")
        self.locate_logout_link(page).click()

        expect(page).to_have_url("/user/logged_out_redirect")
        expect(self.locate_logged_out_alert(page)).to_be_visible()

        expect(self.locate_login_link(page)).to_be_visible()
        expect(self.locate_logout_link(page)).not_to_be_attached()
        milestone_screenshot("post_logout_redirect")

    @pytest.mark.ckan_standard("logout")
    def test_logout_anonymous(self, milestone_screenshot: Any, page: Page):
        """Test anonymous user logout flow.

        Given an anonymous user
        When they go to the logout page
        Then they are redirected to the login page
        """
        page.goto("/user/_logout")
        expect(page).to_have_url("/user/login")
        milestone_screenshot("redirect_to_login")


class Profile(ElementLocator):
    @pytest.mark.ckan_standard("profile")
    def test_profile_tabs(self, milestone_screenshot: Any, login: Any, page: Page, user: dict[str, Any]):
        """Test navigation between profile tabs.

        Given an authenticated user
        When they go to their profile page
        Then they can navigate between the Datasets, Organizations, Groups, and API Tokens tabs
        And the URL updates accordingly
        """
        login(user["name"])

        page.goto("/")
        self.locate_profile_link(page, user).click()
        milestone_screenshot("main_page")

        self.locate_profile_datasets_tab(page).click()
        expect(page).to_have_url(f"/user/{user['name']}")
        milestone_screenshot("datasets_tab")

        self.locate_profile_organizations_tab(page).click()
        expect(page).to_have_url(f"/user/{user['name']}/organizations")
        milestone_screenshot("organizations_tab")

        self.locate_profile_groups_tab(page).click()
        expect(page).to_have_url(f"/user/{user['name']}/groups")
        milestone_screenshot("groups_tab")

        self.locate_profile_tokens_tab(page).click()
        expect(page).to_have_url(f"/user/{user['name']}/api-tokens")
        milestone_screenshot("tokens_tab")

    @pytest.mark.ckan_standard("profile")
    def test_profile_management(
        self, milestone_screenshot: Any, login: Any, page: Page, user: dict[str, Any], faker: Faker
    ):
        """Test profile management flow.

        Given an authenticated user
        When they go to their profile page
        And click the "Manage" link
        And update their display name
        Then they are redirected back to their profile page
        And their new display name is shown in the header
        And the profile link with the old display name is no longer present
        And the user's display name is updated in the database
        """
        login(user["name"])

        page.goto(f"/user/{user['name']}")
        self.locate_profile_management_link(page).click()
        expect(page).to_have_url(f"/user/edit/{user['name']}")

        new_name = faker.name()
        page.get_by_label("full name").fill(new_name)
        milestone_screenshot("form_with_updad_field")

        self.locate_update_profile_button(page).click()
        expect(page).to_have_url(f"/user/{user['name']}")

        expect(self.locate_profile_link(page, user)).not_to_be_attached()

        user = call_action("user_show", id=user["id"])
        assert user["display_name"] == new_name
        expect(self.locate_profile_link(page, user)).to_be_visible()
        milestone_screenshot("updated_profile_page")


class StandardUser(Login, Logout, Profile):
    """Test generic authentication flow."""



from __future__ import annotations

from typing import Any

import pytest

from playwright.sync_api import expect, Page

expect.set_options(timeout=1000)


@pytest.fixture(autouse=True)
def page_timeout_(page: Page):
    """Reduce locator's timeout from 30s to 5s."""
    page.set_default_timeout(5000)


@pytest.fixture
def browser_context_args(
    browser_context_args: dict[str, Any], ckan_config: dict[str, Any]
):
    """Modify playwright's standard configuration of browser's context."""
    browser_context_args["base_url"] = ckan_config["ckan.site_url"]
    return browser_context_args


@pytest.fixture
def login(request: pytest.FixtureRequest, api_token_factory: Any, page: Page):
    """Provides a function for authentication."""

    def authenticator(user: str, _page: Page | None = None):
        if _page is None:
            _page = page

        token: str = api_token_factory(user=user)["token"]
        _page.set_extra_http_headers({"Authorization": token})

    return authenticator


@pytest.fixture
def wait_for_ckan(page: Page):
    """Wait JS initialization before processing with page testing."""
    def waiter(_page: Page | None = None):
        if _page is None:
            _page = page

        page.wait_for_function("() => window.ckan && window.ckan.SITE_ROOT")

    return waiter


@pytest.fixture
def goto(wait_for_ckan: Any, page: Page):
    """Page transition with autowait for CKAN initialization."""

    def switcher(url: str, _page: Page | None = None, **kwargs: Any):
        if _page is None:
            _page = page

        result = page.goto(url, **kwargs)
        wait_for_ckan(page)
        return result

    return switcher



from __future__ import annotations

import logging
from typing import TYPE_CHECKING

import sqlalchemy as sa

import ckan.plugins as p
import ckan.plugins.toolkit as tk
from ckan import model
from ckan.lib.search import commit, rebuild

if TYPE_CHECKING:
    from ckan import types


log = logging.getLogger(__name__)


class CascadeOrganizationUpdatesPlugin(p.SingletonPlugin):
    p.implements(p.IActions)

    def get_actions(self):
        return {
            "organization_update": organization_update,
        }


@tk.chained_action
def organization_update(
    next_: types.Action,
    context: types.Context,
    data_dict: types.ActionResult.OrganizationUpdate,
):
    """Reindex all datasets inside the organization after any update."""
    result = next_(context, data_dict)
    tk.enqueue_job(reindex_organization, [result["id"]])
    return result


def reindex_organization(id_or_name: str):
    """Rebuild search index for all datasets inside the organization."""
    org = model.Group.get(id_or_name)

    if not org:
        log.warning("Organization with ID or name %s not found", id_or_name)
        return

    query = sa.select(model.Package.id).where(model.Package.owner_org == org.id)

    for chunk in model.Session.scalars(query).partitions(100):
        rebuild(package_ids=chunk, force=True, defer_commit=True)

    commit()



"""Views of the {{ cookiecutter.project_shortname }} plugin.

All blueprints added to `__all__` are registered as blueprints inside Flask
app. If you have multiple blueprints, create them inside submodules of
`ckanext.{{ cookiecutter.project_shortname }}.views` and re-export via `__all__`.

Example:
    ```python
    from .custom import custom_bp
    from .data import data_bp

    __all__ = ["custom_bp", "data_bp"]
    ```
"""

from __future__ import annotations

import io
import csv
from typing import Any, cast

from flask import Blueprint, make_response
from flask.views import MethodView

import ckan.plugins.toolkit as tk
from ckan import model
from ckan.logic import parse_params

__all__ = ["bp"]

bp = Blueprint("{{ cookiecutter.project_shortname }}", __name__)


# instead of catching exceptions inside every view, it's usually more
# convenient to register handlers for the exception class.
@bp.errorhandler(tk.ObjectNotFound)
def not_found_handler(error: tk.ObjectNotFound) -> tuple[str, int]:
    """Generic handler for ObjectNotFound exception."""
    return (
        tk.render(
            "error_document_template.html",
            {
                "code": 404,
                "content": f"Object not found: {error.message}",
                "name": "Not found",
            },
        ),
        404,
    )


# error handler renders standard error page. If you want to render
# view-specific page with a flash message instead, it's better it try/catch
# inside the view.
@bp.errorhandler(tk.NotAuthorized)
def not_authorized_handler(error: tk.NotAuthorized) -> tuple[str, int]:
    """Generic handler for NotAuthorized exception."""
    return (
        tk.render(
            "error_document_template.html",
            {
                "code": 403,
                "content": error.message or "Not authorized to view this page",
                "name": "Not authorized",
            },
        ),
        403,
    )


# use `Blueprint.route` decorators to register function-based views. This
# approach is more readable than using `Blueprint.add_url_rule`.
@bp.route("/{{ cookiecutter.project_shortname }}/page")
def page():
    """Basic page."""
    return "Hello, {{ cookiecutter.project_shortname }}!"


@bp.route("/{{ cookiecutter.project_shortname }}/page-redirect")
def page_redirect():
    """Complex page."""
    return tk.redirect_to("{{ cookiecutter.project_shortname }}.page")


# Class-based views cannot be registered via decorator
class ComplexView(MethodView):
    """Complex view.

    Pellentesque dapibus suscipit ligula.  Donec posuere augue in quam.  Etiam
    vel tortor sodales tellus ultricies commodo.  Suspendisse potenti.  Aenean
    in sem ac leo mollis blandit.  Donec neque quam, dignissim in, mollis nec,
    sagittis eu, wisi.  Phasellus lacus.  Etiam laoreet quam sed arcu.
    Phasellus at dui in ligula mollis ultricies.  Integer placerat tristique
    nisl.  Praesent augue.  Fusce commodo.  Vestibulum convallis, lorem a
    tempus semper, dui dui euismod elit, vitae placerat urna tortor vitae
    lacus.  Nullam libero mauris, consequat quis, varius et, dictum id, arcu.
    Mauris mollis tincidunt felis.  Aliquam feugiat tellus ut neque.  Nulla
    facilisis, risus a rhoncus fermentum, tellus tellus lacinia purus, et
    dictum nunc justo sit amet elit.

    Attributes:
        template: page template of the view.
    """

    template = "{{ cookiecutter.project_shortname }}/complex.html"

    def _prepare(self, word: str) -> dict[str, Any]:
        tk.check_access(
            "{{ cookiecutter.project_shortname }}_get_sum",
            {},
            {
                "left": word,
                "right": word,
            },
        )

        username = tk.current_user.name
        if tk.current_user.is_authenticated:
            user = cast(model.User, tk.current_user)
        else:
            user = None

        return {
            "word": word,
            "username": username,
            "user": user,
        }

    def get(self, word: str):
        data = self._prepare(word)
        return tk.render(self.template, data)

    def post(self, word: str):
        data = self._prepare(word)

        params = parse_params(tk.request.form)

        try:
            result = tk.get_action("{{ cookiecutter.project_shortname }}_get_sum")(
                {},
                params,
            )
        except tk.ValidationError as err:
            data["errors"] = err.error_dict
            for field, msg in err.error_summary.items():
                tk.h.flash_error(f"{field}: {msg}")

            return tk.render(self.template, data)

        tk.h.flash_success("Yay! {}".format(result["sum"]))
        return tk.redirect_to("{{ cookiecutter.project_shortname }}.page")


@bp.route("/stats/tracking_data", methods=["GET"])
def export_tracking_data():
    """Exports tracking data as a CSV file.

    Query params:
        start (optional) - filter records from this date (inclusive)
        end (optional) - filter records up to this date (inclusive)
    """
    start_date = tk.request.args.get("start")
    end_date = tk.request.args.get("end")

    query = """
        SELECT
            url,
            tracking_type,
            COUNT(*) AS total_count,
            MIN(access_timestamp) AS first_access,
            MAX(access_timestamp) AS last_access
        FROM tracking_raw
        WHERE url IS NOT NULL AND url <> ''
    """

    # Add date filter if provided
    params = {}
    if start_date:
        query += " AND access_timestamp >= :start_date"
        params["start_date"] = start_date
    if end_date:
        query += " AND access_timestamp <= :end_date"
        params["end_date"] = end_date

    query += " GROUP BY url, tracking_type ORDER BY total_count DESC;"

    results = model.Session.execute(query, params)
    output = io.StringIO()
    writer = csv.writer(output)
    writer.writerow(["url", "tracking_type", "count", "first_access", "last_access"])

    for row in results:
        writer.writerow(row)

    response = make_response(output.getvalue())
    response.headers["Content-Disposition"] = "attachment; filename=page_views.csv"
    response.headers["Content-Type"] = "text/csv"
    return response


# we don't have to specify `methods` parameter, because `MethodView` already
# contains this information
bp.add_url_rule(
    "/{{ cookiecutter.project_shortname }}/complex/<word>",
    view_func=ComplexView.as_view("complex"),
)



"""Create something table.

Revision ID: cc1a832108c5
Revises:
"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects.postgresql import JSONB

# revision identifiers, used by Alembic.
revision = "cc1a832108c5"
down_revision = None
branch_labels = None
depends_on = None


def upgrade():
    op.create_table(
        "{{ cookiecutter.project_shortname }}_something",
        sa.Column("id", sa.UnicodeText, primary_key=True),
        sa.Column("hello", sa.UnicodeText, nullable=False, server_default=""),
        sa.Column("world", sa.UnicodeText, nullable=False),
        sa.Column("plugin_data", JSONB, server_default="{}"),
    )


def downgrade():
    op.drop_table("{{ cookiecutter.project_shortname }}_something")



"""Tests for ckanext.{{ cookiecutter.project_shortname }}.helpers module.

You can either call helpers directly or use `tk.h` and `with_plugins` fixture
to register and call helpers. The former is simpler, the latter verifies that
helpers are registered.
"""

from __future__ import annotations

from ckanext.{{ cookiecutter.project_shortname }} import helpers


def test_hello():
    """Helper returns expected greeting."""
    assert helpers.{{
        cookiecutter.project_shortname
    }}_hello() == "Hello, {{ cookiecutter.project_shortname }}!"



from __future__ import annotations

from typing import Any

from playwright.sync_api import Page


class ElementLocator:
    def locate_login_link(self, page: Page):
        """Locate the "log in" link on the page."""
        return page.get_by_role("link", name="log in")

    def locate_logout_link(self, page: Page):
        """Locate the "log out" link on the page."""
        return page.get_by_label("log out")

    def locate_login_button(self, page: Page):
        """Locate the login button on the login page."""
        return page.get_by_role("button", name="login")

    def locate_profile_link(self, page: Page, user: dict[str, Any]):
        """Locate the profile link on the page based on the user's display name."""
        return page.get_by_text(user["display_name"]).first

    def locate_existing_login_session_alert(self, page: Page):
        """Locate the alert indicating an existing login session."""
        return page.get_by_text("you're already logged in").first

    def locate_logged_out_alert(self, page: Page):
        """Locate the alert indicating the user is logged out."""
        return page.get_by_text("you are now logged out").first

    def locate_main_area(self, page: Page):
        """Locate the main content area of the page."""
        return page.locator("[role=main]")

    def locate_profile_datasets_tab(self, page: Page):
        """Locate the "Datasets" link in the user's profile section."""
        return self.locate_main_area(page).get_by_role("link", name="Datasets")

    def locate_profile_organizations_tab(self, page: Page):
        """Locate the "Organizations" link in the user's profile section."""
        return self.locate_main_area(page).get_by_role("link", name="Organizations")

    def locate_profile_groups_tab(self, page: Page):
        """Locate the "Groups" link in the user's profile section."""
        return self.locate_main_area(page).get_by_role("link", name="Groups")

    def locate_profile_tokens_tab(self, page: Page):
        """Locate the "API Tokens" link in the user's profile section."""
        return self.locate_main_area(page).get_by_role("link", name="API Tokens")

    def locate_profile_management_link(self, page: Page):
        """Locate the "Manage" link in the user's profile section."""
        return self.locate_main_area(page).get_by_role("link", name="Manage")

    def locate_update_profile_button(self, page: Page):
        """Locate the "Update" button on the profile management page."""
        return page.get_by_role("button", name="update")



from __future__ import annotations

from typing import Any, cast

import pytest
from flask_login import encode_cookie  # pyright: ignore[reportUnknownVariableType]
from playwright.sync_api import BrowserContext, Page

from ckan import types

__all__ = ["browser_context_args", "login", "wait_for_ckan", "goto", "ckan_standard", "milestone_screenshot"]


@pytest.fixture(autouse=True)
def ckan_standard(request: pytest.FixtureRequest):
    node = cast(pytest.Function, request.node)
    standard = {feature for marker in node.iter_markers("ckan_standard") for feature in marker.args}
    non_standard = {feature for marker in node.iter_markers("ckan_non_standard") for feature in marker.args}

    conflict = standard & non_standard
    if conflict:
        pytest.skip(f"Non-standard implementation of {conflict}")


@pytest.fixture
def browser_context_args(browser_context_args: dict[str, Any], ckan_config: dict[str, Any]):
    """Modify playwright's standard configuration of browser's context."""
    browser_context_args["base_url"] = ckan_config["ckan.site_url"]
    return browser_context_args


@pytest.fixture
def token_login(api_token_factory: Any, page: Page):
    """Provides a function for authentication using API token."""

    def authenticator(user: str | dict[str, Any], _page: Page | None = None):
        if _page is None:
            _page = page

        if isinstance(user, dict):
            user = user["name"]

        token: str = api_token_factory(user=user)["token"] if user else ""

        _page.set_extra_http_headers({"Authorization": token})

    return authenticator


@pytest.fixture
def login(page: Page, context: BrowserContext, ckan_config: types.FixtureCkanConfig, with_request_context: Any):
    """Provides a function for authentication by setting the remember cookie."""

    def authenticator(user: str | dict[str, Any], _page: Page | None = None):
        if _page is None:
            _page = page

        if isinstance(user, dict):
            user = user["name"]

        key = ckan_config["REMEMBER_COOKIE_NAME"]
        url = ckan_config["ckan.site_url"]

        context.add_cookies([{"name": key, "value": encode_cookie(user), "url": url}])

    return authenticator


@pytest.fixture
def wait_for_ckan(page: Page):
    """Wait for CKAN JS to be initialized.

    Use after page.goto() if you need to interact with widgets that have
    data-module attribute.
    """

    def waiter(_page: Page | None = None):
        if _page is None:
            _page = page

        page.wait_for_function("() => window.ckan && window.ckan.SITE_ROOT")
        page.wait_for_timeout(200)

    return waiter


@pytest.fixture
def goto(wait_for_ckan: Any, page: Page):
    """Page transition with autowait for CKAN initialization."""

    def switcher(url: str, _page: Page | None = None, **kwargs: Any):
        if _page is None:
            _page = page

        result = page.goto(url, **kwargs)
        wait_for_ckan(page)

        return result

    return switcher


@pytest.fixture
def milestone_screenshot(page: Page, request: pytest.FixtureRequest):
    """Fixture to take screenshots at significant steps in a test."""
    step = 1

    def func(name: str, _page: Page | None = None, **kwargs: Any):
        """Takes a screenshot and saves it to the test-results directory."""
        nonlocal step
        node = request.node  # pyright: ignore[reportUnknownVariableType]
        if _page is None:
            _page = page

        prefix: str = node.originalname[5:]  # pyright: ignore[reportUnknownVariableType]
        kwargs["path"] = f"test-results/{prefix}__{step:02d}_{name}.jpeg"
        step += 1
        return _page.screenshot(**kwargs, full_page=True)

    return func



from __future__ import annotations

import copy
from typing import Any

import sqlalchemy as sa
from sqlalchemy.dialects.postgresql import JSONB
from sqlalchemy.orm import Mapped

from ckan.lib.dictization import table_dictize
from ckan.model.types import make_uuid

from .base import Base


class Something(Base):  # type: ignore
    """Model with details or something."""

    # define columns as a `__table__` attribute. It simplifies typing and you
    # can copy this definition almost unchanged into alembic migration.
    __table__ = sa.Table(
        "{{ cookiecutter.project_shortname }}_something",
        Base.metadata,
        sa.Column("id", sa.UnicodeText, primary_key=True, default=make_uuid),
        sa.Column("hello", sa.Text, nullable=False, default=""),
        sa.Column("world", sa.Text, nullable=False),
        sa.Column("plugin_data", JSONB, default=dict, server_default="{}"),
    )

    # typed models. You'll use it - you'll love it.
    id: Mapped[str]

    hello: Mapped[str]
    world: Mapped[str]

    plugin_data: Mapped[dict[str, Any]]

    def dictize(self, context: Any) -> dict[str, Any]:
        result = table_dictize(self, context)

        plugin_data = result.pop("plugin_data")
        if context.get("include_plugin_data"):
            result["plugin_data"] = copy.deepcopy(plugin_data)

        return result

    @classmethod
    def by_hello(cls, hello: str, world: str | None = None) -> sa.sql.Select:
        stmt = sa.select(cls).where(
            cls.hello == hello,
        )

        if world:
            stmt = stmt.where(cls.world == world)

        return stmt



from __future__ import annotations

import click

from . import _shared


@click.group()
def config():
    """Make config files."""


@config.command()
@_shared.option_plugin
@_shared.option_write
def pyproject(plugin: str, write: bool):
    """pyproject.toml."""
    _shared.ensure_root()
    plugin = _shared.safe_plugin_name(plugin)
    data = {
        f"{part.upper()}_CONFIG": _shared.render(
            _shared.template_source(part),
            {"PLUGIN": plugin},
        )
        for part in [
            "black",
            "ruff",
            "isort",
            "pytest",
            "git_changelog",
            "pyright",
            "coverage",
        ]
    }
    data["PLUGIN"] = plugin

    _shared.produce(*_config_files("pyproject"), data, write)


@config.command()
@_shared.option_plugin
@_shared.option_write
def pre_commit(plugin: str, write: bool):
    """Pre-commit (https://pre-commit.com/)."""
    _shared.ensure_root()
    _shared.produce(
        *_config_files("pre-commit"),
        {"PLUGIN": _shared.safe_plugin_name(plugin)},
        write,
    )


@config.command()
@_shared.option_plugin
@_shared.option_write
def ckanext_makefile(plugin: str, write: bool):
    """Tools for CKAN extension management."""
    _shared.ensure_root()
    _shared.produce(
        *_config_files("ckanext-makefile"),
        {"PLUGIN": _shared.safe_plugin_name(plugin)},
        write,
    )


@config.command()
@_shared.option_plugin
@_shared.option_write
def deps_makefile(plugin: str, write: bool):
    """CKAN dependency manager (https://github.com/DataShades/ckan-deps-installer)."""
    _shared.ensure_root()
    _shared.produce(
        *_config_files("deps-makefile"),
        {"PLUGIN": _shared.safe_plugin_name(plugin)},
        write,
    )


@config.command()
@_shared.option_plugin
@_shared.option_write
def gulp_sass(plugin: str, write: bool):
    """Gulpfile for SCSS-based assets."""
    _shared.ensure_root()
    _shared.produce(
        *_config_files("gulp-sass"),
        {"PLUGIN": _shared.safe_plugin_name(plugin)},
        write,
    )


def _config_files(name: str) -> tuple[str, str]:
    if name == "pyproject":
        return f"config_{name}.toml", "pyproject.toml"

    if name == "pre-commit":
        return f"config_{name}.yaml", ".pre-commit-config.yaml"

    if name == "deps-makefile":
        return f"config_{name}.sh", "Makefile"

    if name == "ckanext-makefile":
        return f"config_{name}.sh", "Makefile"

    if name == "gulp-sass":
        return f"config_{name}.js", "gulpfile.js"

    click.secho(f"Unsupported config: {name}", fg="red")
    raise click.Abort



"""Tests for UI."""

from __future__ import annotations

import pytest
from playwright.sync_api import Page

import ckan.plugins.toolkit as tk


@pytest.mark.playwright
def test_homepage(page: Page):
    """Homepage shows about link."""
    page.goto(tk.url_for("home.index"))
    assert page.get_by_text("about").count()



from __future__ import annotations

from typing import Any

import ckan.plugins.toolkit as tk
from ckan.types import Context


@tk.auth_allow_anonymous_access
def {{ cookiecutter.project_shortname }}_get_sum(context: Context, data_dict: dict[str, Any]):
    """Any user can compute sum."""
    return {"success": True}


def {{ cookiecutter.project_shortname }}_something_create(context: Context, data_dict: dict[str, Any]):
    """Authenticated user can create something."""
    return {"success": True}



from __future__ import annotations

import abc
import logging
from typing import TYPE_CHECKING, Any, Iterable

import ckan.plugins.toolkit as tk
from ckan.lib.search.query import solr_literal

from ckanext.toolbelt.utils.structures import Node

if TYPE_CHECKING:
    from typing_extensions import TypeAlias

log = logging.getLogger(__name__)

PackageDict: TypeAlias = "dict[str, Any]"

CONFIG_PARENT_FIELD = "ckanext.toolbelt.package_hierarchy.parent_field"
DEFAULT_PARENT_FIELD = "parent_id"

CONFIG_REFERENCE_FIELD = "ckanext.toolbelt.package_hierarchy.reference_field"
DEFAULT_REFERENCE_FIELD = "id"

CONFIG_PARENT_DISTANCE = "ckanext.toolbelt.package_hierarchy.parent_distance"
DEFAULT_PARENT_DISTANCE = 10

CONFIG_CHILD_DISTANCE = "ckanext.toolbelt.package_hierarchy.child_distance"
DEFAULT_CHILD_DISTANCE = 10

CONFIG_SIBLING_LIMIT = "ckanext.toolbelt.package_hierarchy.sibling_limit"
DEFAULT_SIBLING_LIMIT = 20


def config_parent_field() -> str:
    return tk.config.get(CONFIG_PARENT_FIELD, DEFAULT_PARENT_FIELD)


def config_reference_field() -> str:
    return tk.config.get(CONFIG_REFERENCE_FIELD, DEFAULT_REFERENCE_FIELD)


def config_parent_distance() -> int:
    return tk.asint(tk.config.get(CONFIG_PARENT_DISTANCE, DEFAULT_PARENT_DISTANCE))


def config_child_distance() -> int:
    return tk.asint(tk.config.get(CONFIG_CHILD_DISTANCE, DEFAULT_CHILD_DISTANCE))


def config_sibling_limit() -> int:
    return tk.asint(tk.config.get(CONFIG_SIBLING_LIMIT, DEFAULT_SIBLING_LIMIT))


class Strategy(abc.ABC):
    def __init__(self, context: dict[str, Any]):
        self.context = context

        parent_distance = config_parent_distance()
        child_distance = config_child_distance()
        sibling_limit = config_sibling_limit()

        if tk.request:
            parent_distance = tk.asint(
                tk.request.args.get("__relationship_parent_distance", parent_distance),
            )
            child_distance = tk.asint(
                tk.request.args.get("__relationship_child_distance", child_distance),
            )

            sibling_limit = tk.asint(
                tk.request.args.get("__relationship_sibling_limit", sibling_limit),
            )

        self.parent_distance = parent_distance
        self.child_distance = child_distance
        self.sibling_limit = sibling_limit

    @abc.abstractmethod
    def root(self, pkg: PackageDict) -> tuple[PackageDict, int]:
        raise NotImplementedError

    @abc.abstractmethod
    def children(self, pkg: PackageDict) -> Iterable[PackageDict]:
        raise NotImplementedError


class ParentReference(Strategy):
    def __init__(self, context: dict[str, Any]):
        super().__init__(context)
        self.parent_field = config_parent_field()
        self.reference_field = config_reference_field()

    def root(self, pkg: PackageDict) -> tuple[PackageDict, int]:
        root: PackageDict = pkg
        i = 0
        while i < self.parent_distance:
            if not root.get(self.parent_field):
                break

            key = self.reference_field
            value = solr_literal(root[self.parent_field])
            results = tk.get_action("package_search")(
                dict(self.context),
                {"q": f"{key}:{value}"},
            )["results"]
            if not results:
                break

            if len(results) > 1:
                log.warning(
                    "Multiple packages found by %s with value %s",
                    self.reference_field,
                    root[self.parent_field],
                )
            root = results[0]

            i += 1

        return root, i

    def children(self, pkg: PackageDict):
        key = self.parent_field
        value = pkg[self.reference_field]
        return tk.get_action("package_search")(
            {},
            {
                "fq": f"{key}:({value})",
                "rows": self.sibling_limit,
            },
        )["results"]


def package_hierarchy(
    id_: str,
    context: dict[str, Any] | None = None,
    strategy_factory: type[Strategy] = ParentReference,
) -> Node[PackageDict]:
    """Return a Node starting from the **reacheable** root of the package's
    hierarchy.
    """
    ctx = context or {}
    root: PackageDict = tk.get_action("package_show")(dict(ctx), {"id": id_})

    strategy = strategy_factory(ctx)
    root, distance = strategy.root(root)

    return _package_as_node(root, distance, strategy)


def _package_as_node(
    pkg: PackageDict,
    buffer: int,
    strategy: Strategy,
) -> Node[PackageDict]:
    """Recursive function that wraps package into Node and attach children to
    it as leaves.

    """
    children: Iterable[PackageDict]

    depth_capacity = buffer + strategy.child_distance
    children = strategy.children(pkg) if depth_capacity else []

    node = Node(
        pkg,
        [_package_as_node(child, buffer - 1, strategy) for child in children],
    )

    for child in node.leaves:
        child.parent = node

    return node



from unittest import mock

import pytest

import ckan.plugins.toolkit as tk
from ckan.lib import jobs
from ckan.tests.helpers import RQTestBase, call_action

from ckanext.toolbelt.plugins.cascade_organization_updates import reindex_organization


@pytest.mark.usefixtures("clean_db", "clean_index")
@pytest.mark.ckan_config("ckan.plugins", "toolbelt_cascade_organization_updates")
@pytest.mark.usefixtures("with_plugins")
class TestFlow(RQTestBase):
    def test_basic(self, organization, package_factory):
        pkg = package_factory(owner_org=organization["id"])

        call_action("organization_update", **dict(organization, title="Updated"))

        updated_pkg = call_action("package_show", id=pkg["id"])
        assert updated_pkg["organization"]["title"] != "Updated"

        jobs.Worker().work(True)

        updated_pkg = call_action("package_show", id=pkg["id"])
        assert updated_pkg["organization"]["title"] == "Updated"


@pytest.mark.usefixtures("clean_db")
@pytest.mark.ckan_config("ckan.plugins", "toolbelt_cascade_organization_updates")
@pytest.mark.usefixtures("with_plugins")
class TestEnqueue:
    @mock.patch("ckan.plugins.toolkit.enqueue_job")
    def test_enqueued_with_org_id(self, stub, organization):
        stub.assert_not_called()
        call_action("organization_update", **dict(organization, title="Updated"))

        stub.assert_called_once_with(reindex_organization, [organization["id"]])

    @mock.patch("ckan.plugins.toolkit.enqueue_job")
    def test_not_enqueued_for_missing_organization(self, stub):
        with pytest.raises(tk.ObjectNotFound):
            call_action("organization_update", id="not real")

        stub.assert_not_called()


@pytest.mark.usefixtures("clean_db", "clean_index")
@pytest.mark.ckan_config("ckan.plugins", "toolbelt_cascade_organization_updates")
@pytest.mark.usefixtures("with_plugins")
class TestJob:
    def test_basic(self, organization, package_factory):
        names = ["first", "second"]
        for name in names:
            pkg = package_factory(owner_org=organization["id"], name=name)

        call_action("organization_update", **dict(organization, title="Updated"))

        for name in names:
            pkg = call_action("package_show", id=name)
            assert pkg["organization"]["title"] != "Updated"

        reindex_organization("not-real")

        for name in names:
            pkg = call_action("package_show", id=name)
            assert pkg["organization"]["title"] != "Updated"

        reindex_organization(organization["id"])

        for name in names:
            pkg = call_action("package_show", id=name)
            assert pkg["organization"]["title"] == "Updated"



from __future__ import annotations

import logging
from datetime import datetime
from typing import Any

from flask import Blueprint

import ckan.plugins as p
import ckan.plugins.toolkit as tk
from ckan import model

log = logging.getLogger(__name__)

toolbelt = Blueprint("toolbelt_group_changes", __name__)

CONFIG_WATCH_FIELDS = "ckanext.toolbelt.group_changes.watch_fields"
DEFAULT_WATCH_FIELDS = []


class GroupChangesPlugin(p.SingletonPlugin):
    p.implements(p.IConfigurer, inherit=True)
    p.implements(p.IBlueprint)
    p.implements(p.ITemplateHelpers)

    def get_helpers(self):
        return {
            "compare_group_dicts": compare_group_dicts,
        }

    def get_blueprint(self):
        return [toolbelt]

    def update_config(self, config):
        tk.add_template_directory(config, "templates/group_changes")


def check_metadata_org_changes(change_list, old, new):
    """Compares two versions of a organization and records the changes between
    them in change_list.
    """
    from ckan.lib.changes import _title_change

    # if the title has changed
    if old.get("title") != new.get("title"):
        _title_change(change_list, old, new)

    # if the description of the organization changed
    if old.get("description") != new.get("description"):
        _description_change(change_list, old, new)

    # if the image URL has changed
    if old.get("image_url") != new.get("image_url"):
        _image_url_change(change_list, old, new)

    watched = tk.aslist(tk.config.get(CONFIG_WATCH_FIELDS, DEFAULT_WATCH_FIELDS))
    for field in watched:
        if old.get(field) != new.get(field):
            _custom_field_change(change_list, old, new, field)


def _custom_field_change(
    change_list: list[dict[str, Any]],
    old: dict[str, Any],
    new: dict[str, Any],
    field: str,
):
    """Appends a summary of a change to a organization's description between two
    versions (old and new) to change_list.
    """
    # if the old organization had a description
    if old.get(field) and new.get(field):
        change_list.append(
            {
                "type": field,
                "pkg_id": new.get("id"),
                "title": new.get("title"),
                "new_value": new.get(field),
                "old_value": old.get(field),
                "method": "change",
            },
        )
    elif not new.get(field):
        change_list.append(
            {
                "type": field,
                "pkg_id": new.get("id"),
                "title": new.get("title"),
                "method": "remove",
            },
        )
    else:
        change_list.append(
            {
                "type": field,
                "pkg_id": new.get("id"),
                "title": new.get("title"),
                "new_value": new.get(field),
                "method": "add",
            },
        )


def _description_change(change_list, old, new):
    """Appends a summary of a change to a organization's description between two
    versions (old and new) to change_list.
    """
    # if the old organization had a description
    if old.get("description") and new.get("description"):
        change_list.append(
            {
                "type": "description",
                "pkg_id": new.get("id"),
                "title": new.get("title"),
                "new_description": new.get("description"),
                "old_description": old.get("description"),
                "method": "change",
            },
        )
    elif not new.get("description"):
        change_list.append(
            {
                "type": "description",
                "pkg_id": new.get("id"),
                "title": new.get("title"),
                "method": "remove",
            },
        )
    else:
        change_list.append(
            {
                "type": "description",
                "pkg_id": new.get("id"),
                "title": new.get("title"),
                "new_description": new.get("description"),
                "method": "add",
            },
        )


def _image_url_change(change_list, old, new):
    """Appends a summary of a change to a organization's image URL between two
    versions (old and new) to change_list.
    """
    # if both old and new versions have image  URLs
    if old.get("image_url") and new.get("image_url"):
        change_list.append(
            {
                "type": "image_url",
                "method": "change",
                "pkg_id": new.get("id"),
                "title": new.get("title"),
                "new_image_url": new.get("image_url"),
                "old_image_url": old.get("image_url"),
            },
        )
    # if the user removed the image URL
    elif not new.get("image_url"):
        change_list.append(
            {
                "type": "image_url",
                "method": "remove",
                "pkg_id": new.get("id"),
                "title": new.get("title"),
                "old_image_url": old.get("image_url"),
            },
        )
    # if there wasn't one there before
    else:
        change_list.append(
            {
                "type": "image_url",
                "method": "add",
                "pkg_id": new.get("id"),
                "title": new.get("title"),
                "new_image_url": new.get("image_url"),
            },
        )


def compare_group_dicts(old: dict[str, Any], new: dict[str, Any], old_activity_id: str):
    """Takes two package dictionaries that represent consecutive versions of
    the same organization and returns a list of detailed & formatted summaries
    of the changes between the two versions. old and new are the two package
    dictionaries. The function assumes that both dictionaries will have
    all of the default package dictionary keys, and also checks for fields
    added by extensions and extra fields added by the user in the web
    interface.

    Returns a list of dictionaries, each of which corresponds to a change
    to the dataset made in this revision. The dictionaries each contain a
    string indicating the type of change made as well as other data necessary
    to form a detailed summary of the change.
    """
    change_list: list[dict[str, Any]] = []

    check_metadata_org_changes(change_list, old, new)

    # if the organization was updated but none of the fields we check
    # were changed, display a message stating that
    if len(change_list) == 0:
        change_list.append({"type": "no_change"})

    return change_list


@toolbelt.route("/group/changes/<id>", endpoint="changes")
def changes(id: str) -> str:
    """Shows the changes to an organization in one particular activity stream
    item.
    """
    group_type = "group"
    extra_vars = {}
    activity_id = id
    context = {
        "model": model,
        "session": model.Session,
        "user": tk.g.user,
        "auth_user_obj": tk.g.userobj,
    }
    try:
        activity_diff = tk.get_action("activity_diff")(
            context,
            {
                "id": activity_id,
                "object_type": "group",
                "diff_type": "html",
            },
        )
    except tk.ObjectNotFound as e:
        log.info("Activity not found: %s - %s", e, activity_id)
        return tk.abort(404, tk._("Activity not found"))
    except tk.NotAuthorized:
        return tk.abort(403, tk._("Unauthorized to view activity data"))

    # 'group_dict' needs to go to the templates for page title & breadcrumbs.
    # Use the current version of the package, in case the name/title have
    # changed, and we need a link to it which works
    group_id = activity_diff["activities"][1]["data"]["group"]["id"]
    current_group_dict = tk.get_action(group_type + "_show")(context, {"id": group_id})
    group_activity_list = tk.get_action(group_type + "_activity_list")(
        context,
        {"id": group_id, "limit": 100},
    )

    extra_vars: dict[str, Any] = {
        "activity_diffs": [activity_diff],
        "group_dict": current_group_dict,
        "group_activity_list": group_activity_list,
        "group_type": current_group_dict["type"],
    }

    return tk.render("group/changes.html", extra_vars)


@toolbelt.route("/group_type/changes_multiple", endpoint="changes_multiple")
def changes_multiple() -> str:
    """Called when a user specifies a range of versions they want to look at
    changes between. Verifies that the range is valid and finds the set of
    activity diffs for the changes in the given version range, then
    re-renders changes.html with the list.
    """
    group_type = "group"
    extra_vars = {}
    new_id = tk.h.get_request_param("new_id")
    old_id = tk.h.get_request_param("old_id")

    context = {
        "model": model,
        "session": model.Session,
        "user": tk.g.user,
        "auth_user_obj": tk.g.userobj,
    }

    # check to ensure that the old activity is actually older than
    # the new activity
    old_activity = tk.get_action("activity_show")(
        context,
        {"id": old_id, "include_data": False},
    )
    new_activity = tk.get_action("activity_show")(
        context,
        {"id": new_id, "include_data": False},
    )

    old_timestamp = old_activity["timestamp"]
    new_timestamp = new_activity["timestamp"]

    t1 = datetime.strptime(old_timestamp, "%Y-%m-%dT%H:%M:%S.%f")
    t2 = datetime.strptime(new_timestamp, "%Y-%m-%dT%H:%M:%S.%f")

    time_diff = t2 - t1
    # if the time difference is negative, just return the change that put us
    # at the more recent ID we were just looking at
    # TODO: do something better here - go back to the previous page,
    # display a warning that the user can't look at a sequence where
    # the newest item is older than the oldest one, etc
    if time_diff.total_seconds() < 0:
        return changes(tk.h.get_request_param("current_new_id"))

    done = False
    current_id = new_id
    diff_list = []

    while not done:
        try:
            activity_diff = tk.get_action("activity_diff")(
                context,
                {
                    "id": current_id,
                    "object_type": "group",
                    "diff_type": "html",
                },
            )
        except tk.ObjectNotFound as e:
            log.info("Activity not found: %s - %s", e, current_id)
            return tk.abort(404, tk._("Activity not found"))
        except tk.NotAuthorized:
            return tk.abort(403, tk._("Unauthorized to view activity data"))

        diff_list.append(activity_diff)

        if activity_diff["activities"][0]["id"] == old_id:
            done = True
        else:
            current_id = activity_diff["activities"][0]["id"]

    group_id: str = diff_list[0]["activities"][1]["data"]["group"]["id"]
    current_group_dict = tk.get_action(group_type + "_show")(context, {"id": group_id})
    group_activity_list = tk.get_action(group_type + "_activity_list")(
        context,
        {"id": group_id, "limit": 100},
    )

    extra_vars: dict[str, Any] = {
        "activity_diffs": diff_list,
        "group_dict": current_group_dict,
        "group_activity_list": group_activity_list,
        "group_type": current_group_dict["type"],
    }

    return tk.render("group/changes.html", extra_vars)



from __future__ import annotations

import json
import logging
import pickle  # nosec: B403
from functools import cached_property, update_wrapper
from typing import Any, Callable, Generic, Optional, Protocol, TypeVar, Union, cast

from typing_extensions import ParamSpec

from ckan.lib import redis

from . import constantly

log = logging.getLogger(__name__)
T = TypeVar("T")
TC = TypeVar("TC", covariant=True)
P = ParamSpec("P")


class DontCache(Generic[T]):
    __slots__ = ("value",)
    value: T

    def __init__(self, value: T):
        self.value = value

    def unwrap(self) -> T:
        return self.value


MaybeNotCached = Union[T, DontCache[T]]
CacheAwareCallable = Callable[P, MaybeNotCached[T]]


class NaiveCallable(Generic[P, TC], Protocol):
    def reset(self, *args: P.args, **kwargs: P.kwargs) -> int: ...

    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> TC: ...


CacheDecorator = Callable[[CacheAwareCallable[P, T]], NaiveCallable[P, T]]

Duration = Optional[int]
DurationStrategy = Callable[..., Duration]

KeyStr = Union[str, bytes]
KeyStrategy = Callable[..., KeyStr]

Dumper = Callable[[T], KeyStr]
Loader = Callable[[KeyStr], T]


def default_key_strategy(func: Callable[..., Any], *args: Any, **kwargs: Any) -> bytes:
    return bytes(f"{func.__module__}:{func.__name__}", "utf8") + pickle.dumps(
        (args, kwargs),
    )


def decorate_key_strategy(prefix: bytes, after: bool = False) -> KeyStrategy:
    def strategy(*args: Any, **kwargs: Any):
        left = prefix
        right = default_key_strategy(*args, **kwargs)
        if after:
            left, right = right, left

        return left + right

    return strategy


class Cache(Generic[T]):
    duration: DurationStrategy
    key: KeyStrategy
    dumper: Dumper[T]
    loader: Loader[T]

    def __init__(
        self,
        duration: DurationStrategy | Duration = None,
        key: KeyStrategy | KeyStr = default_key_strategy,
        dumper: Dumper[T] = json.dumps,
        loader: Loader[T] = json.loads,
    ):
        self.dumper = dumper
        self.loader = loader

        if not callable(duration):
            duration = constantly(duration)
        self.duration = duration
        if not callable(key):
            key = constantly(key)
        self.key = key

    @cached_property
    def conn(self):
        return cast("redis.Redis[bytes]", redis.connect_to_redis())

    @staticmethod
    def dont_cache(value: T):
        return DontCache(value)

    def __call__(self, func: CacheAwareCallable[P, T]) -> NaiveCallable[P, T]:
        caller = Caller(func, self)
        return cast(Caller[P, T], update_wrapper(caller, func))


class Caller(NaiveCallable[P, T]):
    def __init__(self, func: CacheAwareCallable[P, T], cache: Cache[T]):
        self.func = func
        self.cache = cache

    def reset(self, *args: P.args, **kwargs: P.kwargs) -> int:
        key = self.cache.key(self.func, *args, **kwargs)
        return self.cache.conn.delete(key)

    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> T:
        key = self.cache.key(self.func, *args, **kwargs)
        old_value = self.cache.conn.get(key)

        if old_value:
            log.debug("Hit cache for key %s", key)
            return self.cache.loader(old_value)

        value = self.func(*args, **kwargs)
        if isinstance(value, DontCache):
            return cast(T, value.unwrap())

        duration = self.cache.duration(self.func, *args, **kwargs)
        self.cache.conn.set(key, self.cache.dumper(value), ex=duration or None)
        return value



"""Tests for ckanext.{{ cookiecutter.project_shortname }}.logic.schema.

If schema does not cover all the fields from user input, don't forget to test
validation inside the action.
"""

from __future__ import annotations

from typing import Any

import pytest

import ckan.plugins.toolkit as tk

from ckanext.{{ cookiecutter.project_shortname }}.logic import schema


@pytest.mark.parametrize(
    ("input", "expected", "failed"),
    [
        ({"left": 1, "right": 2}, {"left": 1, "right": 2}, set()),
        ({"left": 1}, {"left": 1, "right": tk.missing}, {"right"}),
        ({"right": 2}, {"right": 2, "left": tk.missing}, {"left"}),
        ({}, {"left": tk.missing, "right": tk.missing}, {"left", "right"}),
        ({"right": "10", "left": ""}, {"right": 10, "left": ""}, {"left"}),
        (
            {"right": "He", "left": None},
            {"right": "He", "left": None},
            {"right", "left"},
        ),
    ],
)
def test_get_sum(input: dict[str, Any], expected: dict[str, Any], failed: set[str]):
    """Test get_sum schema."""
    data, errors = tk.navl_validate(
        input,
        schema.get_sum(),
        {},
    )

    assert data == expected
    assert set(errors) == failed


@pytest.mark.parametrize(
    ("input", "expected", "failed"),
    [
        ({}, {"hello": tk.missing, "world": tk.missing}, {"hello", "world"}),
        # ... you got the idea
    ],
)
def test_something_create(
    input: dict[str, Any],
    expected: dict[str, Any],
    failed: set[str],
):
    """Test something_create schema."""
    data, errors = tk.navl_validate(
        input,
        schema.something_create(),
        {},
    )

    assert data == expected
    assert set(errors) == failed



"""Config getters of {{ cookiecutter.project_shortname }} plugin."""

from __future__ import annotations

import ckan.plugins.toolkit as tk

OPTION = "ckanext.{{ cookiecutter.project_shortname }}.option.name"
MULTI = "ckanext.{{ cookiecutter.project_shortname }}.multivalued.option"


def option() -> int:
    """Integer placerat tristique nisl."""
    return tk.config[OPTION]


def multivalued() -> list[str]:
    """Another option that will be parsed as a list of words."""
    return tk.config[MULTI]



import pytest
from pytest_mock import MockerFixture

from ckanext.toolbelt.utils import cache


def mul2(v: int):
    return v * 2


def test_dont_cache():
    value = {}
    assert cache.DontCache(value).unwrap() is value


def test_default_key_strategy():
    key = cache.default_key_strategy(mul2, 10)
    assert key.startswith(b"test_cache:mul2")


def test_decoreate_key_strategy():
    key = cache.decorate_key_strategy(b"hello:")(mul2, 10)
    assert isinstance(key, bytes)
    assert key.startswith(b"hello:test_cache:mul2")


@pytest.mark.usefixtures("clean_cache")
class TestCache:
    def fun(self, v):
        return v + v

    def test_basic(self, mocker: MockerFixture):
        spy = mocker.spy(self, "fun")

        decorated = cache.Cache()(self.fun)
        spy.assert_not_called()

        assert decorated(10) == 20
        spy.assert_called_once_with(10)

        assert decorated(20) == 40
        spy.assert_called_with(20)

        assert decorated(10) == 20
        assert spy.call_count == 2

        decorated.reset(10)
        assert decorated(10) == 20
        assert spy.call_count == 3

        decorated.reset(20)
        assert decorated(10) == 20
        assert spy.call_count == 3



"""Tests for ckanext.{{ cookiecutter.project_shortname }}.view.

Here you define functional tests that verify page rendering process. It's also
possible to test views using cypress.

Pytest performs better when you want to something on server side befor the
assertion. Cypress should be preferred when you only test user interaction with
frontend without executing arbitrary python code.
"""

import bs4
import pytest
from faker import Faker

import ckan.plugins.toolkit as tk
from ckan.tests.helpers import CKANTestApp


# blueprints are registered when plugin is enabled, so `with_plugins` is
# mandatory fixture for any view test.
@pytest.mark.usefixtures("with_plugins")
def test_page(app: CKANTestApp):
    """Basic page is rendered without errors."""
    resp = app.get(tk.url_for("{{ cookiecutter.project_shortname }}.page"))
    assert resp.status_code == 200
    assert resp.body == "Hello, {{ cookiecutter.project_shortname }}!"


@pytest.mark.usefixtures("with_plugins")
def test_page_redirect(app: CKANTestApp):
    """Basic page is rendered without errors."""
    # application follows redirects automatically. In order to check presence
    # of `Location` header, you need to disable this default behavior.
    resp = app.get(
        tk.url_for("{{ cookiecutter.project_shortname }}.page_redirect"),
        follow_redirects=False,
    )

    assert resp.status_code == 302

    # `Location` header contains absolute URL to the target page
    expected = tk.url_for("{{ cookiecutter.project_shortname }}.page", _external=True)
    assert resp.headers["location"] == expected


@pytest.mark.usefixtures("with_plugins")
class TestComplex:
    def test_get(self, app: CKANTestApp, faker: Faker):
        """Complex page is rendered with expected content."""
        word = faker.word()
        resp = app.get(tk.url_for("{{ cookiecutter.project_shortname }}.complex", word=word))
        # BeautifulSoup parses the response content and produces searchable DOM
        # tree.
        page = bs4.BeautifulSoup(resp.body)

        # select element from the page using CSS selector. Selecting elements
        # by tag name is not wise, but we know that page contains exactly one
        # heading.
        heading = page.select_one("h1")
        assert heading
        assert heading.text == f"Hello, {word}!"

    def test_post_invalid(self, app: CKANTestApp, faker: Faker):
        """Complex page shows an error when arguments are missing."""
        word = faker.word()
        resp = app.post(tk.url_for("{{ cookiecutter.project_shortname }}.complex", word=word))
        page = bs4.BeautifulSoup(resp.body)

        # `.select` selects all elements matching selector, while `.select_one`
        # returns only the first.
        errors = {x.text.strip() for x in page.select(".flash-messages .alert")}

        # use set to avoid errors because of unknown order of messages
        assert errors == {
            "Left: Missing value",
            "Right: Missing value",
        }

    def test_post_valid(self, app: CKANTestApp, faker: Faker):
        """Complex page shows no errors for valid requests."""
        word = faker.word()
        resp = app.post(
            tk.url_for("{{ cookiecutter.project_shortname }}.complex", word=word),
            data={"left": "1", "right": "2"},
        )
        page = bs4.BeautifulSoup(resp.body)

        assert page.select(".flash-messages .alert") == []



"""Tests for ckanext.{{ cookiecutter.project_shortname }}.logic.validators."""

import pytest

import ckan.plugins.toolkit as tk

from ckanext.{{ cookiecutter.project_shortname }}.logic import validators


def test_required_with_valid_value():
    """Non-empty value is accepted."""
    assert validators.{{
        cookiecutter.project_shortname}}_required("value") == "value"


def test_required_with_invalid_value():
    """Missing value is not accepted."""
    with pytest.raises(tk.Invalid):
        validators.{{ cookiecutter.project_shortname }}_required(None)


def test_complex():
    """Do something complex here."""
    key = ("name",)
    errors = {key: []}

    with pytest.raises(tk.StopOnError):
        validators.{{ cookiecutter.project_shortname }}_complex_validator(
            key,
            {key: tk.missing},
            errors,
            {},
        )

    assert errors[key] == ["Required"]
